@@ -301,7 +301,7 @@ errcode_t ext2fs_flush2(ext2_filsys fs, int flags)
 	fs_state = fs->super->s_state;
 	feature_incompat = fs->super->s_feature_incompat;
 
-	fs->super->s_wtime = fs->now ? fs->now : time(NULL);
+	ext2fs_set_tstamp(fs->super, s_wtime, ext2fsP_get_time(fs));
 	fs->super->s_block_group_nr = 0;
 
 	/*
