@@ -29,18 +29,19 @@
 int main(int argc, const char *argv[])
 {
 	errcode_t	retval;
-	blk_t		blocks;
+	blk64_t		blocks;
 
 	if (argc < 2) {
 		fprintf(stderr, "%s device\n", argv[0]);
 		exit(1);
 	}
 	add_error_table(&et_ext2_error_table);
-	retval = ext2fs_get_device_size(argv[1], 1024, &blocks);
+	retval = ext2fs_get_device_size2(argv[1], 1024, &blocks);
 	if (retval) {
 		com_err(argv[0], retval, "while getting device size");
 		exit(1);
 	}
-	printf("%s is device has %u blocks.\n", argv[1], blocks);
+	printf("%s is device has %llu blocks.\n", argv[1],
+	       (unsigned long long) blocks);
 	return 0;
 }
