@@ -20,8 +20,7 @@
 
 #include <stdarg.h>
 
-char * ss_name(sci_idx)
-    int sci_idx;
+char *ss_name(int sci_idx)
 {
     register char *ret_val;
     register ss_data *infop;
@@ -43,6 +42,8 @@ char * ss_name(sci_idx)
 			 (strlen(infop->subsystem_name)+
 			  strlen(infop->current_request)+
 			  4));
+	if (ret_val == (char *)NULL)
+		return ((char *)NULL);
 	cp = ret_val;
 	cp1 = infop->subsystem_name;
 	while (*cp1)
@@ -70,10 +71,7 @@ void ss_error (int sci_idx, long code, const char * fmt, ...)
     va_end(pvar);
 }
 
-void ss_perror (sci_idx, code, msg) /* for compatibility */
-    int sci_idx;
-    long code;
-    char const *msg;
+void ss_perror(int sci_idx, long code, char const *msg) /* for compatibility */
 {
     ss_error (sci_idx, code, "%s", msg);
 }
