@@ -25,7 +25,14 @@ my @translator_help = (
  "#. it can expand to either the string \"indirect block\" (possibly preceded\n",
  "#. by the word \"double\" or \"triple\"), or the string \"block #\" immediately\n",
  "#. followed by an integer indicating a block sequence number.\n",
- "#.  \n",
+ "#.\n",
+ "#. Please note that the %-expansion for most e2fsck's problem.c should not\n",
+ "#. use positional indicators such as %1$c, since although they look like c-style\n",
+ "#. format strings, they are NOT c-style format strings, and the positional\n",
+ "#. indicators (which BTW are GNU extensions and so won't work on other Unix\n",
+ "#. gettext implementations) won't work with e2fsck's print_e2fsck_message()\n",
+ "#. function found in e2fsck/message.c\n",
+ "#.\n",
  "#.	%b	<blk>			block number\n",
  "#.	%B	\"indirect block\" | \"block #\"<blkcount>	string | string+integer\n",
  "#.	%c	<blk2>			block number\n",
